unit HTTPDownloader;

interface

uses
  Windows, Messages, SysUtils, Classes, ScktComp, WinSock;


type
  TRange = record AFrom,ATo:longint; end;
  TInformer = procedure (Completed,Total:longint; data:string) of object;
  TInformer2 = procedure (Info:string) of object;
  TProblem = procedure (kind:integer) of object;
    {-----------------------------------
     Kind description:
      1 = Invalid HTTP response
      2 = Couldn't determine file size
      3 = Resume not supported
      4 = Incomplete Download
      5 = 404 Not Found
     -----------------------------------}
  THTTPDownloader = class(TClientSocket)
    constructor Create(Aowner:Tcomponent); override;
    destructor Destroy; override;
   private
    FOncompl,FHRead:TInformer2;
    GSize,FCompl,Size:longint;
    FCustomParams,FPOSTStrings:tstrings;
    DoPost,FLHost,FRSupp,FPaused,FFinished,HeaderLoaded,FKAlive:boolean;
    FRecvd: TInformer;
    FRange: TRange;
    FName,FURL,FReferer,FUAgent,FRequestedFile,Header,DFN: string;
    FProblem: TProblem;
    procedure SetURL(s:string);
    procedure TDError(Sender: TObject;
      Socket: TCustomWinSocket; ErrorEvent: TErrorEvent;
      var ErrorCode: Integer);
    procedure DConnect(Sender: TObject; Socket: TCustomWinSocket);
    procedure DDisconnect(Sender: TObject; Socket: TCustomWinSocket);
    procedure DRead(Sender: TObject; Socket: TCustomWinSocket);
    function PrepareRequest:string;
    function CheckForHeader:boolean;
    property Active;
    function GetIAddr(s: string): string;
    function GetPath(s: string): string;
    function IsIP(s: string): boolean;
    procedure SetStrings1(a:TStrings);
    procedure SetStrings2(a:TStrings);
    { Private declarations }
  protected
    { Protected declarations }
  public

     FStream: TFileStream;
     property DStream:TFileStream read fstream write fstream;
     procedure Run;
     procedure Pause;
     procedure Stop;
    { Public declarations }
  published
     property POST:boolean read DoPost write DoPost;
     property CustomParams:TStrings read FCustomParams write SetStrings1;
     property POSTStrings:TStrings read FPostStrings write SetStrings2;
     property ForceLocalHost:boolean read FLHost write FLHost;
     property ResumeSupported:boolean read FRSupp;
     property Paused:boolean read FPaused;
     property Completed:integer read FCompl;
     property Finished:boolean read ffinished;
     property URL:string read furl write seturl;
     property Port;
     property KeepAlive:boolean read fkalive write fkalive;
     property DestName:string read dfn write dfn;
     property Range:trange read frange write frange;
     property Referer:string read freferer write freferer;
     property UserAgent:string read fUAgent write fUAgent;
     property OnProblem:TProblem read fproblem write fproblem;
     property OnComplete:Tinformer2 read FOnCompl write FOnCompl;
     property OnPacketRecvd:TInformer read frecvd write frecvd;
     property OnHeaderRead:TInformer2 read FHRead write FHRead;
    { Published declarations }
  end;
  function drange(a,b:longint):TRange;
procedure Register;

implementation

function drange(a,b:longint):TRange;
var
r:trange;
begin
r.AFrom:=a;
r.ATo:=b;
result:=r;
end;

procedure Register;
begin
  RegisterComponents('Standard', [THTTPDownloader]);
end;

{ THTTPDownloader }

procedure THTTPDownloader.SetStrings1(a:TStrings);
begin
FCustomParams.Text:=a.text;
end;

procedure THTTPDownloader.SetStrings2(a:TStrings);
begin
FPOSTStrings.Text:=a.text;
end;

function THTTPDownloader.CheckForHeader:boolean;
label 1;
var
  errorcode:integer;
  ex:tstringlist;
  x:integer;
  s:string;
begin
errorcode:=0;
ex:=tstringlist.Create;
ex.Text:=header;
frsupp:=true;
if pos('404 Not Found', ex.Strings[0])>0 then begin errorcode:=5; goto 1; end;
if (range.AFrom>-1)and(range.ATo>-1) then if pos('206', ex.Strings[0])=0 then begin frsupp:=false; errorcode:=3; goto 1; end;
if range.afrom=-1 then if pos('200',ex.Strings[0])=0 then begin errorcode:=1; goto 1; end;

for x:=1 to ex.Count-1 do
  if pos('Content-Length: ',ex.Strings[x])=1 then begin
    s:=ex.strings[x];
    delete(s,1,length('Content-Length: '));
    try
      size:=strtoint(s);
      if gsize=0 then gsize:=size;
    except size:=-1; end;
  end;

if range.afrom>-1 then if size=-1 then begin frsupp:=false; errorcode:=2; goto 1; end;

if assigned(OnHeaderRead) then OnHeaderRead(ex.Text);
ex.free;
result:=true;
exit;
1:
if assigned(OnHeaderRead) then OnHeaderRead(ex.Text);
if assigned(onproblem) then onproblem(errorcode);
ex.free;
result:=false;
end;

constructor THTTPDownloader.Create(Aowner: Tcomponent);
begin
inherited create(aowner);
DoPost:=false;
FPOSTStrings:=tstringlist.Create;
FCustomParams:=tstringlist.Create;
flhost:=false;
OnError:=tderror;
OnConnect:=DConnect;
FUAgent:='Mozilla/4.0 (compatible; MSIE 5.00; Windows 98)';
frange:=drange(-1,-1);
port:=80;
KeepAlive:=true;
FFinished:=false;
FCompl:=0;
GSize:=0;
FPaused:=false;
FRSupp:=false;
end;

procedure THTTPDownloader.DConnect(Sender: TObject;
  Socket: TCustomWinSocket);
var
s:string;
begin
socket.Lock;
s:=PrepareRequest;
socket.SendText(s);
end;

procedure THTTPDownloader.DDisconnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
socket.Unlock;
if Paused then exit;
if completed<gsize then onProblem(4);
Range:=drange(-1,-1);
FFinished:=true;
if assigned(OnComplete) then OnComplete('Complete');
if fstream <> nil then fstream.Free;
fcompl:=0;
gsize:=0;
end;

destructor THTTPDownloader.Destroy;
begin
active:=false;
FPOSTStrings.Destroy;
FCustomParams.Destroy;
inherited destroy;
end;

procedure THTTPDownloader.DRead(Sender: TObject; Socket: TCustomWinSocket);
var
s,s2:string;
b:pchar;
q:integer;
begin
Socket.ReceiveLength;
s:=socket.ReceiveText;
if not HeaderLoaded then
begin
  s2:=Header+s;
  q:=pos(#13#10#13#10,s2);
  if q=0 then
  begin
    Header:=s2;
    s:='';
  end else
  begin
    Header:=copy(s2,1,q);
    delete(s,1,q+3);
    HeaderLoaded:=CheckForHeader;
    if not HeaderLoaded then
      begin
        HeaderLoaded:=true;
      end;
  end;
end;

q:=length(s);
  if q>0 then
  begin
    b:=pchar(s);
    fstream.Write(b^,q);
    FCompl:=fCompl+q;
    if assigned(OnPacketRecvd) then OnPacketRecvd(fcompl,gsize,s);
    if gsize>0 then
    if fcompl>=gsize then active:=false;
  end;
end;

function THTTPDownloader.PrepareRequest: string;
var
ex:tstringlist;
x:integer;
s:string;
begin
ex:=tstringlist.Create;
if not POST then
  ex.Add('GET /'+fRequestedFile+' HTTP/1.1') else
  ex.Add('POST /'+fRequestedFile+' HTTP/1.1');
ex.add('HOST: '+Host);
ex.add('ACCEPT: */*');
if referer='' then referer:='http://'+host;
ex.add('Referer: '+referer);
ex.Add('User-Agent: '+useragent);
if range.AFrom>-1 then
if range.ATo>-1 then
    ex.Add('Range: bytes='+inttostr(range.AFrom)+'-'+inttostr(range.ato))
  else
    ex.Add('Range: bytes='+inttostr(range.afrom)+'-');
ex.Add('Pragma: no-cache');
ex.Add('Cache-Control: no-cache');
if keepalive then
    ex.Add('Connection: Keep-Alive')
  else
    ex.Add('Connection: close');


for x:=0 to CustomParams.Count-1 do
if CustomParams.Strings[x]<>'' then ex.Add(CustomParams.Strings[x]);

if POST then
  begin
    s:='';
    for x:=0 to POSTStrings.Count-1 do
    if PostStrings.Strings[x]<>'' then s:=s+'&'+PostStrings.Strings[x];
    if length(s)>0 then if s[1]='&' then delete(s,1,1);
    ex.Add('Content-length: '+inttostr(length(s)));
  end;
ex.add('');
if POST then ex.Add(s);
result:=ex.Text;
ex.free;
end;

procedure THTTPDownloader.Run;
begin
if not paused then
  begin
    if dfn='' then dfn:=fname;
    if dfn='' then dfn:='index.html';
    if not fileexists(dfn) then
      begin
        fstream:=tfilestream.Create(dfn,fmcreate);
        fstream.Free;
      end;
    fstream:=tfilestream.Create(dfn,fmOpenReadWrite+fmsharedenynone);
    fstream.Position:=0;
    if range.AFrom>=0 then
      begin
        if fstream.Size<range.AFrom then fstream.Size:=range.AFrom;
        fstream.Position:=range.AFrom;
      end;
  end else FPaused:=false;
size:=-1;
header:='';
headerloaded:=false;
onread:=dread;
ondisconnect:=ddisconnect;
FFinished:=false;
Active:=true;
end;

function THTTPDownloader.GetIAddr(s:string):string;
var
e:string;
begin
e:=copy(s,8,length(s));
result:=copy(e,1,pos('/',e)-1);
end;

function THTTPDownloader.GetPath(s:string):string;
var
e:string;
begin
e:=copy(s,8,length(s));
result:=copy(e,pos('/',e)+1,length(e));
end;

function THTTPDownloader.IsIP(s:string):boolean;
var
e:string;
x:integer;
begin
result:=false;
if length(s)<8 then exit;
e:=s;
for x:=1 to length(e) do
if ((e[x]<'0')or(e[x]>'9'))and(e[x]<>'.') then exit;
x:=pos('.',e);
if (x<2) or (x>4) then exit;
e:=copy(e,x+1,length(e));
x:=pos('.',e);
if (x<2) or (x>4) then exit;
e:=copy(e,x+1,length(e));
x:=pos('.',e);
if (x<2) or (x>4) then exit;
e:=copy(e,x+1,length(e));
if length(e)>3 then exit;
result:=true;
end;

procedure THTTPDownloader.SetURL(s: string);
var
s1,s2,s4:string;
x:integer;
begin
FURL:=s;
x:=length(s);
repeat
x:=x-1;
until s[x]='/';
s1:=copy(s,1,x);
s2:=copy(s,x+1,length(s));
s4:=getiaddr(s1);
Address:='';
Host:='';
if isip(s4) then
Address:=s4 else
Host:=s4;
if forcelocalhost then
  begin
    Address:='127.0.0.1';
    host:='localhost';
  end;
fname:=s2;
fRequestedFile:=getpath(s1)+s2;
end;

procedure THTTPDownloader.Stop;
begin
if paused then
  begin
    Range:=drange(-1,-1);
    FFinished:=true;
    fstream.Free;
    fpaused:=false;
  end;
gsize:=0;
fcompl:=0;
active:=false;
end;

procedure THTTPDownloader.TDError(Sender: TObject;
  Socket: TCustomWinSocket; ErrorEvent: TErrorEvent;
  var ErrorCode: Integer);
begin
headerloaded:=false;
if assigned(onproblem) then OnProblem(errorcode);
errorcode:=0;
end;

procedure THTTPDownloader.Pause;
begin
if not ResumeSupported then exit;
fpaused:=true;
active:=false;
Range:=drange(Completed,range.ATo);
end;


end.
